/*
 * Copyright (c) 2011, Noel Lemouel
 * All rights reserved.
 */

#define _WIN32_WINNT 0x0400

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

#include <conio.h>

#ifdef __cplusplus
extern "C" {
#endif

#include <windows.h>
#include <process.h>

#ifdef __cplusplus
} /* extern "C" */
#endif

#include "serial.h"

#define VERSION __DATE__

#define EOT 0x04

static unsigned int timeout = 0x00;

/* needs to be accessed/freed outside of main after CTRL+C */
static unsigned char* bin = NULL;
static unsigned int dev_num;

HANDLE hEvent[2] = {NULL, NULL};
HANDLE hThread = NULL;
HANDLE hStdin = NULL;

/* forward */
static int sc_open(FILE**, const char*, struct stat*);

unsigned __stdcall TF(void* arg) {

   HANDLE* hEvent;

   hEvent = (HANDLE*)arg;

   // WaitForSingleObject(hEvent[0], INFINITE);
   WaitForMultipleObjects(2, hEvent, 0, INFINITE);

   timeout = 1;

   // CloseHandle(hEvent[0]);
   // CloseHandle(hEvent[1]);
   _endthreadex(0);
}

int download(unsigned int base, unsigned int size, unsigned char* bin, unsigned char type)
{
   int i, j, k;
   unsigned char rc[4];
   unsigned char buf[128];

   /* send the command */

   // | D | L | a3 | a2 | a1 | a0 | - | s3 | s2 | s1 | s0 | - | - | - | - | DT |
   // DT = 'C' => code
   // DT = 'D' => data

   buf[0] = 'D';
   buf[1] = 'L';
   buf[2] = (base & 0xff000000) >> 24;
   buf[3] = (base & 0x00ff0000) >> 16;
   buf[4] = (base & 0x0000ff00) >> 8;
   buf[5] = base & 0x000000ff;
   buf[6] = 0xaa;
   buf[7] = (size & 0xff000000) >> 24;
   buf[8] = (size & 0x00ff0000) >> 16;
   buf[9] = (size & 0x0000ff00) >> 8;
   buf[10] = size & 0x000000ff;
   buf[11] = 0xaa;
   buf[12] = 0xaa;
   buf[13] = 0xaa;
   buf[14] = 0xaa;
   buf[15] = type;

   serial_tx(dev_num, buf, 16);

   /* create timer (timeout) */

   timeout = 0;

   hEvent[0] = CreateWaitableTimer(
      0,
      0,     /* 0 => will automatically reset */
      0);    /* name */

   LARGE_INTEGER li;

   li.QuadPart = -(1*(10*1000*1000));    /* 1 seconds [100 ns intervals] */

   SetWaitableTimer(
      hEvent[0],
      &li,    /* first time delay */
      0,      /* run once only */
      0,
      0,
      0);

   if (!hEvent[0]) {
      printf("dl_error: SetWaitableTimer error\n");
   }

   hEvent[1] = CreateEvent (NULL, 0, 0, 0);

   hThread = (HANDLE)_beginthreadex(NULL,0,TF,(void*)hEvent,0,0);
   if (!hThread) {
      printf("dl_error: _beginthreadex error [%d]\n", errno);
      CloseHandle(hEvent[0]);
   }

   /* read the answer */

   k = 0;
   for(;;)
   {
      j = serial_rx(dev_num, buf, 128);
      if(j)
      {
         i = 0;
         while(j-- && k<4) rc[k] = buf[i++], k++;
      }
      if(k == 4) break;
      if(timeout)
      {
         printf("\ndl_error: timeout\n");
         CloseHandle(hEvent[0]);
         CloseHandle(hEvent[1]);
         CloseHandle(hThread);
         return -1;
      }
      usleep(10000);
   }

   /* no timeout */
   CancelWaitableTimer(hEvent[0]);
   SetEvent (hEvent[1]);   /* terminate thread */

   CloseHandle(hEvent[0]);
   CloseHandle(hEvent[1]);
   CloseHandle(hThread);

   /* ... */

   printf("dl |0x%02x|0x%02x|0x%02x|0x%02x|\n", rc[0], rc[1], rc[2], rc[3]);

   if(strncmp("ACK", (char*)rc, 4))
   {
      printf("\ndl_error: invalid response\n");
      return -1;
   }

   printf(" \n");

   /* verify ACK... */

   printf("uploading data... ");

   /* upload data */

   serial_tx(dev_num, bin, size);

   printf("done");

   return 0;
}

/* expects 0xdeadbeef or deadbeef */
unsigned int atox(unsigned char *ip_str)
{
   unsigned int op_val=0, i=0, ip_len = strlen(ip_str);

   if(strncmp(ip_str, "0x", 2) == 0){
      ip_str +=2;
      ip_len -=2;
   }

   for(i=0;i<ip_len;i++){
      op_val *= 0x10;
      switch(ip_str[i]){
         case 'A':
         case 'a':
            op_val += 0xa;
         break;
         case 'B':
         case 'b':
            op_val += 0xb;
         break;
         case 'C':
         case 'c':
            op_val += 0xc;
         break;
         case 'D':
         case 'd':
            op_val += 0xd;
         break;
         case 'E':
         case 'e':
            op_val += 0xe;
         break;
         case 'F':
         case 'f':
            op_val += 0xf;
         break;
         case '0' ... '9':
            op_val += (ip_str[i] - 0x30);
         break;
         default :
            op_val += 0x0;
         break;
      }
   }

   return op_val;
}

static void _cleanup(void)
{
      /* HANDLEs */
      if(hThread)
      {
         /* also check the timeout var. */
         /* just checking hThread should be enough */
         if(!timeout) SetEvent (hEvent[1]);   /* terminate thread */
         CloseHandle(hThread);
      }
      if(hEvent[0]) CloseHandle(hEvent[0]);
      if(hEvent[1]) CloseHandle(hEvent[1]);
      if(hStdin) CloseHandle(hStdin);
      /* memory */
      if(bin) free(bin);
      /* serial IO */
      serial_close(dev_num);
}

static BOOL WINAPI ctrl_handler(DWORD dwCtrlType)
{
   switch (dwCtrlType) {
   case CTRL_C_EVENT:
   case CTRL_BREAK_EVENT:
      _cleanup();
      /* ... */
      exit(0);
   }

   exit(-1);
}

int main(int argc, char *argv[])
{
   int i, j;
   int term;
   int baudrate;
   unsigned char c;
   unsigned char rx[512];
   unsigned char* bin;

   FILE* fd;
   struct stat info;

   DWORD dwStdinMode;

   /* check arguments */

   term = 0;         /* no terminal */
   baudrate = 9600;  /* default baudrate */

   if((argc == 2) && !strcmp("-v", argv[1]))
   {
      printf("WIN32 %s\n", VERSION);
      return 0;
   }

   if((argc >= 5) && (argc % 2)){

      if( sc_open(&fd, argv[2], &info) ) return -1;

      for(i=5; i<argc-1; i++)    /* last argv can't be an opt */
      {
         if( !strcmp("-t", argv[i]) )
         {
            /* printf("TTY [%s] ignored [win32 version]\n\n", argv[i+1]); */ term = 1;
         }
         if( !strcmp("-b", argv[i]) )
         {
            baudrate = atoi(argv[i+1]); // printf("baudrate = %d\n", baudrate);
         }
      }
   }else{
      printf("usage: %s PORT# *.bin 0xXXXXXXXX 0xXX [-b baudrate] [-t tty]\n", basename(argv[0]));
      return -1;
   }

   /* ... */

   printf("-- ORX serial loader --\n\n");

   dev_num = atoi(argv[1]);

   /* load bin file */

   bin = malloc(info.st_size);
   if(!bin){
      printf("error: [alloc. %d] failed\n", (int)info.st_size);
      fclose(fd);
      return -1;
   }else{
      // printf("[alloc. %d]\n", (int)info.st_size);
   }

   fread(bin, (int)info.st_size, 1, fd);   

   fclose(fd);

   if ( !SetConsoleCtrlHandler(ctrl_handler, TRUE) ) {
      printf("error: SetConsoleCtrlHandler error [%d]\n", errno);
      exit(-1);
   }

   /* ... */

   if(serial_open(dev_num, baudrate)) return -1;

   /* ... */

   printf("port  : COM%d [%d 8N1]\n", atoi(argv[1]), baudrate);
   printf("type  : 0x%02x [ignored]\n", atox(argv[4]));
   printf("size  : 0x%08x\n", info.st_size);
   printf("offset: 0x%08x\n", atox(argv[3]));

   if( download(atox(argv[3]), info.st_size, bin, 'C') )
   {
      free(bin);
      serial_close(dev_num);
      return -1;
   }

   if(term)
   {
      /* remove ctrl_handler() HandlerRoutine */
      SetConsoleCtrlHandler(ctrl_handler, FALSE);
      /* ignore CTRL+C input */
      SetConsoleCtrlHandler(NULL, TRUE);
      /* unprocessed input / CTRL+C now is now placed in the input buffer */
      hStdin = GetStdHandle(STD_INPUT_HANDLE);
      if(!hStdin){
         printf("terminal mode error\n");
         _cleanup();
         return(-1);
      }
      if (!GetConsoleMode(hStdin, &dwStdinMode)){
         printf("terminal mode error\n");
         _cleanup();
         return(-1);
      }
      if (!SetConsoleMode(hStdin, dwStdinMode & (~ENABLE_PROCESSED_INPUT))){
         printf("terminal mode error\n");
         _cleanup();
         return(-1);
      }

      printf("\n\n[terminal mode] CTRL+D to quit\n\n");

      while(term)
      {
         i = serial_rx(dev_num, rx, 512);
         if(i)
         {
            for(j=0; j<i; j++)
            {
               if(rx[j]==EOT)
               {
                  term=0;  /* EOT received */
                  break;
               }
               putchar(rx[j]);
            }
         }
         if(_kbhit())
         {
            c = _getch();
            if(c==EOT) break;
            serial_tx(dev_num, &c, 1);
         }
         usleep(10000);
      }
   }

   printf("\n");

   _cleanup();
   
   return 0;
}

static int sc_open(FILE** fd, const char* name, struct stat* info)
{
   if (stat(name, info) == -1)
   {
      if (errno == ENOENT)
      {
         (void) fprintf( stderr, "error: '%s' does not exist\r\n", name );
      }else{
         (void) fprintf( stderr, "error: stat() fails for '%s'\r\n", name );
      }
      return -1;
   }

   /* checking S_ISLNK(m) is not required since 'stat' follows sym. links */
   if (!S_ISREG(info->st_mode))
   {
      (void) fprintf( stderr, "error: '%s' is a not a regular file (or a sym. link)\r\n", name);
      return -1;
   }

   if ((*fd = fopen(name, "rb")) == NULL)
   {
      (void) fprintf( stderr, "error: '%s' open failed\r\n", name );
      return -1;
   }

   /* printf("size: %d\n", (int)info->st_size); */

   return 0;   /* no error */
}

